# file mostly from SciPy:
# https://github.com/scipy/scipy/blob/914523af3bc03fe7bf61f621363fca27e97ca1d6/scipy/optimize/nonlin.py#L221
# and converted to PyTorch for GPU efficiency

from typing import Callable
import warnings
import torch
import functools
from deepchem.utils.differentiation_utils.optimize.jacobian import BroydenFirst, \
    BroydenSecond, LinearMixing
from deepchem.utils import ConvergenceWarning


def _nonlin_solver(
        fcn: Callable,
        x0: torch.Tensor,
        params,
        method: str,
        # jacobian parameters
        alpha=None,
        uv0=None,
        max_rank=None,
        # stopping criteria
        maxiter=None,
        f_tol=None,
        f_rtol=None,
        x_tol=None,
        x_rtol=None,
        # algorithm parameters
        line_search=True,
        # misc parameters
        verbose=False,
        custom_terminator=None,
        **unused):
    """
    Parameters
    ----------
    alpha: float or None
        The initial guess of inverse Jacobian is ``- alpha * I + u v^T``.
    uv0: tuple of tensors or str or None
        The initial guess of inverse Jacobian is ``- alpha * I + u v^T``.
        If ``"svd"``, then it uses 1-rank svd to obtain ``u`` and ``v``.
        If None, then ``u`` and ``v`` are zeros.
    max_rank: int or None
        The maximum rank of inverse Jacobian approximation. If ``None``, it
        is ``inf``.
    maxiter: int or None
        Maximum number of iterations, or inf if it is set to None.
    f_tol: float or None
        The absolute tolerance of the norm of the output ``f``.
    f_rtol: float or None
        The relative tolerance of the norm of the output ``f``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.
    line_search: bool or str
        Options to perform line search. If ``True``, it is set to ``"armijo"``.
    verbose: bool
        Options for verbosity
    """

    if method == "broyden1":
        jacobian = BroydenFirst(alpha=alpha, uv0=uv0, max_rank=max_rank)
    elif method == "broyden2":
        jacobian = BroydenSecond(alpha=alpha, uv0=uv0, max_rank=max_rank)
    elif method == "linearmixing":
        jacobian = LinearMixing(alpha=alpha)  # type: ignore
    else:
        raise RuntimeError("Unknown method: %s" % method)

    if maxiter is None:
        maxiter = 100 * (torch.numel(x0) + 1)
    if line_search is True:
        line_search = "armijo"
    elif line_search is False:
        line_search = None

    # solving complex rootfinder by concatenating real and imaginary part,
    # making the variable twice as long
    x_is_complex = torch.is_complex(x0)

    def _ravel(x: torch.Tensor) -> torch.Tensor:
        # represents x as a long real vector
        if x_is_complex:
            return torch.cat((x.real, x.imag), dim=0).reshape(-1)
        else:
            return x.reshape(-1)

    def _pack(x: torch.Tensor) -> torch.Tensor:
        # pack a long real vector into the shape accepted by fcn
        if x_is_complex:
            n = len(x) // 2
            xreal, ximag = x[:n], x[n:]
            x = xreal + 1j * ximag
        return x.reshape(xshape)

    # shorthand for the function
    xshape = x0.shape

    def func(x):
        return _ravel(fcn(_pack(x), *params))

    x = _ravel(x0)

    y = func(x)
    y_norm = y.norm()
    stop_cond = custom_terminator if custom_terminator is not None \
        else TerminationCondition(f_tol, f_rtol, y_norm, x_tol, x_rtol)
    if (y_norm == 0):
        return x.reshape(xshape)

    # set up the jacobian
    jacobian.setup(x, y, func)

    # solver tolerance
    gamma = 0.9
    eta_max = 0.9999
    eta_threshold = 0.1
    eta = 1e-3

    converge = False
    best_ynorm = y_norm
    best_x = x
    best_dxnorm = x.norm()
    best_iter = 0
    for i in range(maxiter):
        tol = min(eta, eta * y_norm)
        dx = -jacobian.solve(y, tol=tol)

        dx_norm = dx.norm()
        if dx_norm == 0:
            raise ValueError("Jacobian inversion yielded zero vector. "
                             "This indicates a bug in the Jacobian "
                             "approximation.")

        if line_search:
            s, xnew, ynew, y_norm_new = _nonline_line_search(
                func, x, y, dx, search_type=line_search)
        else:
            xnew = x + dx
            ynew = func(xnew)
            y_norm_new = ynew.norm()

        # save the best results
        if y_norm_new < best_ynorm:
            best_x = xnew
            best_dxnorm = dx_norm
            best_ynorm = y_norm_new
            best_iter = i + 1

        jacobian.update(xnew.clone(), ynew)

        # print out dx and df
        to_stop = stop_cond.check(xnew, ynew, dx)
        if verbose:
            if i < 10 or i % 10 == 0 or to_stop:
                print("%6d: |dx|=%.3e, |f|=%.3e" % (i, dx_norm, y_norm))
        if to_stop:
            converge = True
            break

        # adjust forcing parameters for inexact solve
        eta_A = float(gamma * (y_norm_new / y_norm)**2)
        gamma_eta2 = gamma * eta * eta
        if gamma_eta2 < eta_threshold:
            eta = min(eta_max, eta_A)
        else:
            eta = min(eta_max, max(eta_A, gamma_eta2))

        y_norm = y_norm_new
        x = xnew
        y = ynew
    if not converge:
        msg = ("The rootfinder does not converge after %d iterations. "
               "Best |dx|=%.3e, |f|=%.3e at iter %d") % (maxiter, best_dxnorm,
                                                         best_ynorm, best_iter)
        warnings.warn(ConvergenceWarning(msg))
        x = best_x
    return _pack(x)


@functools.wraps(_nonlin_solver,
                 assigned=('__annotations__',))  # takes only the signature
def broyden1(fcn: Callable, x0: torch.Tensor, params=(), **kwargs):
    """
    Solve the root finder or linear equation using the first Broyden method [1]_.
    It can be used to solve minimization by finding the root of the
    function's gradient.

    Examples
    --------
    >>> def fcn(x):
    ...    return x**2 - 4
    >>> x0 = torch.tensor(0.0, requires_grad=True)
    >>> x = broyden1(fcn, x0)
    >>> x
    tensor(-2.0000, grad_fn=<ViewBackward0>)

    Parameters
    ----------
    fcn: callable
        The function to solve. It should take a tensor and return a tensor.
    x0: torch.Tensor
        The initial guess of the solution.
    params: tuple
        The parameters to pass to the function.

    References
    ----------
    .. [1] B.A. van der Rotten, PhD thesis,
           "A limited memory Broyden method to solve high-dimensional systems of nonlinear equations".
           Mathematisch Instituut, Universiteit Leiden, The Netherlands (2003).
           https://web.archive.org/web/20161022015821/http://www.math.leidenuniv.nl/scripties/Rotten.pdf
    .. [2] https://github.com/xitorch/xitorch

    """
    return _nonlin_solver(fcn, x0, params, "broyden1", **kwargs)


@functools.wraps(_nonlin_solver,
                 assigned=('__annotations__',))  # takes only the signature
def broyden2(fcn: Callable, x0: torch.Tensor, params=(), **kwargs):
    """
    Solve the root finder or linear equation using the second Broyden method [2]_.
    It can be used to solve minimization by finding the root of the
    function's gradient.

    Examples
    --------
    >>> def fcn(x):
    ...    return x**2 - 4
    >>> x0 = torch.tensor(0.0, requires_grad=True)
    >>> x = broyden1(fcn, x0)
    >>> x
    tensor(-2.0000, grad_fn=<ViewBackward0>)

    Parameters
    ----------
    fcn: callable
        The function to solve. It should take a tensor and return a tensor.
    x0: torch.Tensor
        The initial guess of the solution.
    params: tuple
        The parameters to pass to the function.

    References
    ----------
    .. [1] B.A. van der Rotten, PhD thesis,
           "A limited memory Broyden method to solve high-dimensional systems of nonlinear equations".
           Mathematisch Instituut, Universiteit Leiden, The Netherlands (2003).
           https://web.archive.org/web/20161022015821/http://www.math.leidenuniv.nl/scripties/Rotten.pdf
    .. [2] https://github.com/xitorch/xitorch
    """
    return _nonlin_solver(fcn, x0, params, "broyden2", **kwargs)


def linearmixing(
        fcn: Callable,
        x0: torch.Tensor,
        params=(),
        # jacobian parameters
        alpha=None,
        # stopping criteria
        maxiter=None,
        f_tol=None,
        f_rtol=None,
        x_tol=None,
        x_rtol=None,
        # algorithm parameters
        line_search=True,
        # misc parameters
        verbose=False,
        **unused):
    """
    Solve the root finding problem by approximating the inverse of Jacobian
    to be a constant scalar.

    Examples
    --------
    >>> def fcn(x):
    ...    return x**2 - 4
    >>> x0 = torch.tensor(0.0, requires_grad=True)
    >>> x = broyden1(fcn, x0)
    >>> x
    tensor(-2.0000, grad_fn=<ViewBackward0>)

    Parameters
    ----------
    fcn: Callable
        The function to solve. It should take a tensor and return a tensor.
    x0: torch.Tensor
        The initial guess of the solution.
    params: tuple
        The parameters to pass to the function.
    alpha: float or None
        The initial guess of inverse Jacobian is ``-alpha * I``.
    maxiter: int or None
        Maximum number of iterations, or inf if it is set to None.
    f_tol: float or None
        The absolute tolerance of the norm of the output ``f``.
    f_rtol: float or None
        The relative tolerance of the norm of the output ``f``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.
    line_search: bool or str
        Options to perform line search. If ``True``, it is set to ``"armijo"``.
    verbose: bool
        Options for verbosity

    References
    ----------
    .. [1] https://github.com/xitorch/xitorch

    """
    kwargs = {
        "alpha": alpha,
        "maxiter": maxiter,
        "f_tol": f_tol,
        "x_tol": x_tol,
        "x_rtol": x_rtol,
        "line_search": line_search,
        "verbose": verbose,
        **unused
    }
    return _nonlin_solver(fcn, x0, params, "linearmixing", **kwargs)


def _safe_norm(v):
    """Compute the norm of a vector, checking for finite values."""
    if not torch.isfinite(v).all():
        return torch.tensor(float("inf"), dtype=v.dtype, device=v.device)
    return torch.norm(v)


def _nonline_line_search(func: Callable,
                         x: torch.Tensor,
                         y: torch.Tensor,
                         dx: torch.Tensor,
                         search_type="armijo",
                         rdiff=1e-8,
                         smin=1e-2):
    """Find a suitable step length for a line search.

    Parameters
    ----------
    func: Callable
        The function to minimize.
    x: torch.Tensor
        The current point.
    y: torch.Tensor
        The function value at the current point.
    dx: torch.Tensor
        The search direction.
    search_type: str
        The type of line search to perform. Currently, only "armijo" is supported.
    rdiff: float
        The relative difference to compute the derivative.
    smin: float
        The minimum step length to take.

    Returns
    -------
    s: float
        The step length.
    x: torch.Tensor
        The new point.
    y: torch.Tensor
        The function value at the new point.
    y_norm: float
        The norm of the function value at the new point.

    """
    tmp_s = [0]
    tmp_y = [y]
    tmp_phi = [y.norm()**2]
    s_norm = x.norm() / dx.norm()

    def phi(s, store=True):
        if s == tmp_s[0]:
            return tmp_phi[0]
        xt = x + s * dx
        v = func(xt)
        p = _safe_norm(v)**2
        if store:
            tmp_s[0] = s
            tmp_phi[0] = p
            tmp_y[0] = v
        return p

    def derphi(s):
        ds = (torch.abs(s) + s_norm + 1) * rdiff
        return (phi(s + ds, store=False) - phi(s)) / ds

    if search_type == 'armijo':
        s, phi1 = _scalar_search_armijo(phi, tmp_phi[0], -tmp_phi[0], amin=smin)

    if s is None:
        # No suitable step length found. Take the full Newton step,
        # and hope for the best.
        s = 1.0

    x = x + s * dx
    if s == tmp_s[0]:
        y = tmp_y[0]
    else:
        y = func(x)
    y_norm = y.norm()

    return s, x, y, y_norm


def _scalar_search_armijo(phi: Callable,
                          phi0: float,
                          derphi0: float,
                          c1: float = 1e-4,
                          alpha0=1,
                          amin=0,
                          max_niter=20):
    """Minimize over alpha, the function phi(s) at the current point and
    the derivative derphi(s) at the current point.

    Parameters
    ----------
    phi: callable
        The function to minimize.
    phi0: float
        The value of phi at 0.
    derphi0: float
        The value of the derivative of phi at 0.
    c1: float
        The Armijo condition parameter.
    alpha0: float
        The initial guess of the step length.
    amin: float
        The minimum step length to take.
    max_niter: int
        The maximum number of iterations to take.

    Returns
    -------
    alpha: float
        The step length.
    phi: float
        The value of the function at the step length.

    """
    phi_a0 = phi(alpha0)
    if phi_a0 <= phi0 + c1 * alpha0 * derphi0:
        return alpha0, phi_a0

    # Otherwise, compute the minimizer of a quadratic interpolant:

    alpha1 = -(derphi0) * alpha0**2 / 2.0 / (phi_a0 - phi0 - derphi0 * alpha0)
    phi_a1 = phi(alpha1)

    if (phi_a1 <= phi0 + c1 * alpha1 * derphi0):
        return alpha1, phi_a1

    # Otherwise, loop with cubic interpolation until we find an alpha which
    # satisfies the first Wolfe condition (since we are backtracking, we will
    # assume that the value of alpha is not too small and satisfies the second
    # condition.
    niter = 0
    while alpha1 > amin and niter < max_niter:  # we are assuming alpha>0 is a descent direction
        factor = alpha0**2 * alpha1**2 * (alpha1 - alpha0)
        a = alpha0**2 * (phi_a1 - phi0 - derphi0 * alpha1) - \
            alpha1**2 * (phi_a0 - phi0 - derphi0 * alpha0)
        a = a / factor
        b = -alpha0**3 * (phi_a1 - phi0 - derphi0 * alpha1) + \
            alpha1**3 * (phi_a0 - phi0 - derphi0 * alpha0)
        b = b / factor

        alpha2 = (-b + torch.sqrt(torch.abs(b**2 - 3 * a * derphi0))) / (3.0 *
                                                                         a)
        phi_a2 = phi(alpha2)

        if (phi_a2 <= phi0 + c1 * alpha2 * derphi0):
            return alpha2, phi_a2

        if (alpha1 - alpha2) > alpha1 / 2.0 or (1 - alpha2 / alpha1) < 0.96:
            alpha2 = alpha1 / 2.0

        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi_a2
        niter += 1

    # Failed to find a suitable step length
    if niter == max_niter:
        return alpha2, phi_a2
    return None, phi_a1


class TerminationCondition(object):
    """Class to check the termination condition of the root finder."""

    def __init__(self, f_tol: float, f_rtol: float, f0_norm: float,
                 x_tol: float, x_rtol: float):
        """Initialize the termination condition.

        Parameters
        ----------
        f_tol: float
            The absolute tolerance of the norm of the output ``f``.
        f_rtol: float
            The relative tolerance of the norm of the output ``f``.
        f0_norm: float
            The norm of the initial function value.
        x_tol: float
            The absolute tolerance of the norm of the input ``x``.
        x_rtol: float
            The relative tolerance of the norm of the input ``x``.

        """
        if f_tol is None:
            f_tol = 1e-6
        if f_rtol is None:
            f_rtol = float("inf")
        if x_tol is None:
            x_tol = 1e-6
        if x_rtol is None:
            x_rtol = float("inf")
        self.f_tol = f_tol
        self.f_rtol = f_rtol
        self.x_tol = x_tol
        self.x_rtol = x_rtol
        self.f0_norm = f0_norm

    def check(self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor) -> bool:
        """Check the termination condition.

        Parameters
        ----------
        x: torch.Tensor
            The current point.
        y: torch.Tensor
            The function value at the current point.
        dx: torch.Tensor
            The search direction.

        Returns
        -------
        bool
            Whether the termination condition is met.

        """
        xnorm = x.norm()
        ynorm = y.norm()
        dxnorm = dx.norm()
        xtcheck = dxnorm < self.x_tol
        xrcheck = dxnorm < self.x_rtol * xnorm
        ytcheck = ynorm < self.f_tol
        yrcheck = ynorm < self.f_rtol * self.f0_norm
        return xtcheck and xrcheck and ytcheck and yrcheck



from typing import List

import torch.nn as nn
import torch

from deepchem.models.losses import L2Loss
from deepchem.models.torch_models import layers
from deepchem.models.torch_models import TorchModel
from deepchem.data.datasets import Dataset
from deepchem.utils import batch_coulomb_matrix_features


class DTNN(nn.Module):
    """Deep Tensor Neural Networks

    DTNN is based on the many-body Hamiltonian concept, which is a fundamental principle in quantum mechanics.
    The DTNN recieves a molecule's distance matrix and membership of its atom from its Coulomb Matrix representation.
    Then, it iteratively refines the representation of each atom by considering its interactions with neighboring atoms.
    Finally, it predicts the energy of the molecule by summing up the energies of the individual atoms.

    In this class, we establish a sequential model for the Deep Tensor Neural Network (DTNN) [1]_.

    Examples
    --------
    >>> import os
    >>> import torch
    >>> from deepchem.models.torch_models import DTNN
    >>> from deepchem.data import SDFLoader
    >>> from deepchem.feat import CoulombMatrix
    >>> from deepchem.utils import batch_coulomb_matrix_features
    >>> # Get Data
    >>> model_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    >>> dataset_file = os.path.join(model_dir, 'tests/assets/qm9_mini.sdf')
    >>> TASKS = ["alpha", "homo"]
    >>> loader = SDFLoader(tasks=TASKS, featurizer=CoulombMatrix(29), sanitize=True)
    >>> data = loader.create_dataset(dataset_file, shard_size=100)
    >>> inputs = batch_coulomb_matrix_features(data.X)
    >>> atom_number, distance, atom_membership, distance_membership_i, distance_membership_j = inputs
    >>> inputs = [torch.tensor(atom_number).to(torch.int64),
    ...           torch.tensor(distance).to(torch.float32),
    ...           torch.tensor(atom_membership).to(torch.int64),
    ...           torch.tensor(distance_membership_i).to(torch.int64),
    ...           torch.tensor(distance_membership_j).to(torch.int64)]
    >>> n_tasks = data.y.shape[0]
    >>> model = DTNN(n_tasks)
    >>> pred = model(inputs)

    References
    ----------
    .. [1] Schütt, Kristof T., et al. "Quantum-chemical insights from deep
        tensor neural networks." Nature communications 8.1 (2017): 1-8.

    """

    def __init__(self,
                 n_tasks: int,
                 n_embedding: int = 30,
                 n_hidden: int = 100,
                 n_distance: int = 100,
                 distance_min: float = -1,
                 distance_max: float = 18,
                 output_activation: bool = True,
                 mode: str = "regression",
                 dropout: float = 0.0,
                 n_steps: int = 2):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_embedding: int (default 30)
            Number of features per atom.
        n_hidden: int (default 100)
            Number of features for each molecule after DTNNStep
        n_distance: int (default 100)
            granularity of distance matrix
            step size will be (distance_max-distance_min)/n_distance
        distance_min: float (default -1)
            minimum distance of atom pairs (in Angstrom)
        distance_max: float (default 18)
            maximum distance of atom pairs (in Angstrom)
        output_activation: bool (default True)
            determines whether an activation function should be apply  to its output.
        mode: str (default "regression")
            Only "regression" is currently supported.
        dropout: float (default 0.0)
            the dropout probablity to use.
        n_steps: int (default 2)
            Number of DTNNStep Layers to use.

        """
        super(DTNN, self).__init__()
        self.n_tasks = n_tasks
        self.n_embedding = n_embedding
        self.n_hidden = n_hidden
        self.n_distance = n_distance
        self.distance_min = distance_min
        self.distance_max = distance_max
        self.output_activation = output_activation
        self.mode = mode
        self.dropout = dropout
        self.n_steps = n_steps

        # get DTNNEmbedding
        self.dtnn_embedding = layers.DTNNEmbedding(n_embedding=self.n_embedding)

        # get DTNNSteps
        self.dtnn_step = nn.ModuleList()
        for i in range(self.n_steps):
            self.dtnn_step.append(
                layers.DTNNStep(n_embedding=self.n_embedding,
                                n_distance=self.n_distance))

        # get DTNNGather
        self.dtnn_gather = layers.DTNNGather(
            n_embedding=self.n_embedding,
            layer_sizes=[self.n_hidden],
            n_outputs=self.n_tasks,
            output_activation=self.output_activation)

        # get Final Linear Layer
        self.linear = nn.LazyLinear(self.n_tasks)

    def forward(self, inputs: List[torch.Tensor]):
        """
        Parameters
        ----------
        inputs: List
            A list of tensors containing atom_number, distance,
            atom_membership, distance_membership_i, and distance_membership_j.

        Returns
        -------
        output: torch.Tensor
            Predictions of the Molecular Energy.

        """
        dtnn_embedding = self.dtnn_embedding(inputs[0])

        for i in range(self.n_steps):
            dtnn_embedding = nn.Dropout(self.dropout)(dtnn_embedding)
            dtnn_embedding = self.dtnn_step[i](
                [dtnn_embedding, inputs[1], inputs[3], inputs[4]])

        dtnn_step = nn.Dropout(self.dropout)(dtnn_embedding)
        dtnn_gather = self.dtnn_gather([dtnn_step, inputs[2]])

        dtnn_gather = nn.Dropout(self.dropout)(dtnn_gather)
        output = self.linear(dtnn_gather)
        return output


class DTNNModel(TorchModel):
    """Implements DTNN models for regression.

    DTNN is based on the many-body Hamiltonian concept, which is a fundamental principle in quantum mechanics.
    DTNN recieves a molecule's distance matrix and membership of its atom from its Coulomb Matrix representation.
    Then, it iteratively refines the representation of each atom by considering its interactions with neighboring atoms.
    Finally, it predicts the energy of the molecule by summing up the energies of the individual atoms.

    This class implements the Deep Tensor Neural Network (DTNN) [1]_.

    Examples
    --------
    >>> import os
    >>> from deepchem.data import SDFLoader
    >>> from deepchem.feat import CoulombMatrix
    >>> from deepchem.models.torch_models import DTNNModel
    >>> model_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    >>> dataset_file = os.path.join(model_dir, 'tests/assets/qm9_mini.sdf')
    >>> TASKS = ["alpha", "homo"]
    >>> loader = SDFLoader(tasks=TASKS, featurizer=CoulombMatrix(29), sanitize=True)
    >>> data = loader.create_dataset(dataset_file, shard_size=100)
    >>> n_tasks = data.y.shape[1]
    >>> model = DTNNModel(n_tasks,
    ...                   n_embedding=20,
    ...                   n_distance=100,
    ...                   learning_rate=1.0,
    ...                   mode="regression")
    >>> loss = model.fit(data, nb_epoch=250)
    >>> pred = model.predict(data)

    References
    ----------
    .. [1] Schütt, Kristof T., et al. "Quantum-chemical insights from deep
        tensor neural networks." Nature communications 8.1 (2017): 1-8.

    """

    def __init__(self,
                 n_tasks: int,
                 n_embedding: int = 30,
                 n_hidden: int = 100,
                 n_distance: int = 100,
                 distance_min: float = -1,
                 distance_max: float = 18,
                 output_activation: bool = True,
                 mode: str = "regression",
                 dropout: float = 0.0,
                 n_steps: int = 2,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_embedding: int (default 30)
            Number of features per atom.
        n_hidden: int (default 100)
            Number of features for each molecule after DTNNStep
        n_distance: int (default 100)
            granularity of distance matrix
            step size will be (distance_max-distance_min)/n_distance
        distance_min: float (default -1)
            minimum distance of atom pairs (in Angstrom)
        distance_max: float (default = 18)
            maximum distance of atom pairs (in Angstrom)
        output_activation: bool (default True)
            determines whether an activation function should be apply  to its output.
        mode: str (default "regression")
            Only "regression" is currently supported.
        dropout: float (default 0.0)
            the dropout probablity to use.
        n_steps: int (default 2)
            Number of DTNNStep Layers to use.

        """
        if dropout < 0 or dropout > 1:
            raise ValueError("dropout probability has to be between 0 and 1, "
                             "but got {}".format(dropout))
        model = DTNN(n_tasks=n_tasks,
                     n_embedding=n_embedding,
                     n_hidden=n_hidden,
                     n_distance=n_distance,
                     distance_min=distance_min,
                     distance_max=distance_max,
                     output_activation=output_activation,
                     mode=mode,
                     dropout=dropout,
                     n_steps=n_steps)
        if mode not in ['regression']:
            raise ValueError("Only 'regression' mode is currently supported")
        super(DTNNModel, self).__init__(model, L2Loss(), ["prediction"],
                                        **kwargs)

    def default_generator(self,
                          dataset: Dataset,
                          epochs: int = 1,
                          mode: str = 'fit',
                          deterministic: bool = True,
                          pad_batches: bool = True):
        """Create a generator that iterates batches for a dataset.
        It processes inputs through the _compute_features_on_batch function to calculate required features of input.

        Parameters
        ----------
        dataset: Dataset
            the data to iterate
        epochs: int
            the number of times to iterate over the full dataset
        mode: str
            allowed values are 'fit' (called during training), 'predict' (called
            during prediction), and 'uncertainty' (called during uncertainty
            prediction)
        deterministic: bool
            whether to iterate over the dataset in order, or randomly shuffle the
            data for each epoch
        pad_batches: bool
            whether to pad each batch up to this model's preferred batch size

        Returns
        -------
        a generator that iterates batches, each represented as a tuple of lists:
        ([inputs], [outputs], [weights])

        """
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                yield (batch_coulomb_matrix_features(X_b,
                                                     self.model.distance_max,
                                                     self.model.distance_min,
                                                     self.model.n_distance),
                       [y_b], [w_b])



"""
EditableModule Utils
Derived From: https://github.com/xitorch/xitorch/blob/master/xitorch/_core/editable_module.py
"""
import inspect
import warnings
from abc import abstractmethod
import copy
import torch
from typing import Sequence, Union, Dict, List, Callable, Any
from deepchem.utils.attribute_utils import get_attr, set_attr, del_attr

__all__ = ["EditableModule"]

torch_float_type = [torch.float32, torch.float, torch.float64, torch.float16]


class EditableModule(object):
    """EditableModule is a base class to enable classes that it inherits be
    converted to pure functions for higher order derivatives purpose.

    Usage
    -----
    To use this class, the user must implement the ``getparamnames`` method
    which returns a list of tensor names that affect the output of the method
    with name indicated in ``methodname``.

    Used in:

    - Classes of Density Functional Theory (DFT).

    - It can also be used in other classes that need to be converted to pure
      functions for higher order derivatives purpose.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import EditableModule
    >>> class A(EditableModule):
    ...     def __init__(self, a):
    ...         self.b = a*a
    ...
    ...     def mult(self, x):
    ...         return self.b * x
    ...
    ...     def getparamnames(self, methodname, prefix=""):
    ...         if methodname == "mult":
    ...             return [prefix+"b"]
    ...         else:
    ...             raise KeyError()
    >>> a = torch.tensor(2.0).requires_grad_()
    >>> x = torch.tensor(0.4).requires_grad_()
    >>> alpha = A(a)
    >>> alpha.mult(x)
    tensor(1.6000, grad_fn=<MulBackward0>)
    >>> alpha.getparamnames("mult")
    ['b']
    >>> alpha.assertparams(alpha.mult, x)
    "mult" method check done

    """

    def getparams(self, methodname: str) -> Sequence[torch.Tensor]:
        """Returns a list of tensor parameters used in the object's operations.
        Requires the ``getparamnames`` method to be implemented.

        Parameters
        ----------
        methodname: str
            The name of the method of the class.

        Returns
        -------
        Sequence[torch.Tensor]
            Sequence of tensors that are involved in the specified method of the
            object.

        """

        paramnames = self.cached_getparamnames(methodname)
        return [get_attr(self, name) for name in paramnames]

    def setparams(self, methodname: str, *params) -> int:
        """Set the input parameters to the object's parameters to make a copy of
        the operations.

        Parameters
        ----------
        methodname: str
            The name of the method of the class.
        *params:
            The parameters to be set to the object's parameters.

        Returns
        -------
        int
            The number of parameters that are set to the object's parameters.

        """
        paramnames = self.cached_getparamnames(methodname)
        for name, val in zip(paramnames, params):
            try:
                set_attr(self, name, val)
            except TypeError:  # failed because val should be param
                del_attr(self, name)
                set_attr(self, name, val)

        return len(params)

    def cached_getparamnames(self,
                             methodname: str,
                             refresh: bool = False) -> List[str]:
        """getparamnames, but cached, so it is only called once

        Parameters
        ----------
        methodname: str
            The name of the method of the class.
        refresh: bool
            If True, the cache is refreshed.

        Returns
        -------
        List[str]
            Sequence of name of parameters affecting the output of the method.

        """
        if not hasattr(self, "_paramnames_"):
            self._paramnames_: Dict[str, List[str]] = {}

        if methodname not in self._paramnames_:
            self._paramnames_[methodname] = self.getparamnames(methodname)
        return self._paramnames_[methodname]

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        This method should list tensor names that affect the output of the
        method with name indicated in ``methodname``.
        If the ``methodname`` is not on the list in this function, it should
        raise ``KeyError``.

        Parameters
        ---------
        methodname: str
            The name of the method of the class.
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            Sequence of name of parameters affecting the output of the method.

        Raises
        ------
        KeyError
            If the list in this function does not contain ``methodname``.

        """
        pass

    def getuniqueparams(self,
                        methodname: str,
                        onlyleaves: bool = False) -> List[torch.Tensor]:
        """Returns the list of unique parameters involved in the method
        specified by `methodname`.

        Parameters
        ----------
        methodname: str
            Name of the method where the returned parameters play roles.
        onlyleaves: bool
            If True, only returns leaf tensors. Otherwise, returns all tensors.

        Returns
        -------
        List[torch.Tensor]
            List of tensors that are involved in the specified method of the
            object.

        """
        allparams = self.getparams(methodname)
        idxs = self._get_unique_params_idxs(methodname, allparams)
        if onlyleaves:
            return [allparams[i] for i in idxs if allparams[i].is_leaf]
        else:
            return [allparams[i] for i in idxs]

    def setuniqueparams(self, methodname: str, *uniqueparams) -> int:
        """Set the input parameters to the object's parameters to make a copy of
        the operations. The input parameters are unique parameters, i.e. they
        are not necessarily the same tensors as the object's parameters.

        Note: This function can only be run after running getuniqueparams.

        Parameters
        ----------
        methodname: str
            The name of the method of the class.
        *uniqueparams:
            The parameters to be set to the object's parameters. The number of
            parameters must be the same as the number of unique parameters
            returned by ``getuniqueparams``.

        Returns
        -------
        int
            The number of parameters that are set to the object's parameters.

        """
        nparams = self._number_of_params[methodname]
        allparams = [None for _ in range(nparams)]
        maps = self._unique_params_maps[methodname]

        for j in range(len(uniqueparams)):
            jmap = maps[j]
            p = uniqueparams[j]
            for i in jmap:
                allparams[i] = p

        return self.setparams(methodname, *allparams)

    def _get_unique_params_idxs(
            self,
            methodname: str,
            allparams: Union[Sequence[torch.Tensor],
                             None] = None) -> Sequence[int]:
        """Returns the list of unique parameters involved in the method

        Parameters
        ----------
        methodname: str
            Name of the method where the returned parameters play roles.
        allparams: list of tensors
            List of tensors that are involved in the specified method of the
            object.

        Returns
        -------
        Sequence[int]
            List of indices of the unique parameters in the list of all parameters.

        """

        if not hasattr(self, "_unique_params_idxs"):
            self._unique_params_idxs = {}  # type: Dict[str,Sequence[int]]
            self._unique_params_maps = {}
            self._number_of_params = {}

        if methodname in self._unique_params_idxs:
            return self._unique_params_idxs[methodname]
        if allparams is None:
            allparams = self.getparams(methodname)

        # get the unique ids
        ids = []  # type: List[int]
        idxs = []
        idx_map = []  # type: List[List[int]]
        for i in range(len(allparams)):
            param = allparams[i]
            id_param = id(param)

            # search the id if it has been added to the list
            try:
                jfound = ids.index(id_param)
                idx_map[jfound].append(i)
                continue
            except ValueError:
                pass

            ids.append(id_param)
            idxs.append(i)
            idx_map.append([i])

        self._number_of_params[methodname] = len(allparams)
        self._unique_params_idxs[methodname] = idxs
        self._unique_params_maps[methodname] = idx_map
        return idxs

    # debugging
    def assertparams(self, method: Callable, *args, **kwargs):
        """
        Perform a rigorous check on the implemented ``getparamnames``
        in the class for a given method and its parameters as well as keyword
        Parameters.
        It raises warnings if there are missing or excess parameters in the
        ``getparamnames`` implementation.

        Parameters
        ---------
        method: Callable
            The method of this class to be tested
        *args:
            Parameters of the method
        **kwargs:
            Keyword parameters of the method

        """
        # check the method input
        if not inspect.ismethod(method):
            raise TypeError("The input method must be a method")
        methodself = method.__self__
        if methodself is not self:
            raise RuntimeError(
                "The method does not belong to the same instance")

        methodname = method.__name__

        # assert if the method preserve the float tensors of the object
        self.__assert_method_preserve(method, *args, **kwargs)
        self.__assert_get_correct_params(
            method, *args,
            **kwargs)  # check if getparams returns the correct tensors
        print('"%s" method check done' % methodname)

    def __assert_method_preserve(self, method, *args, **kwargs):
        """This method assert if method does not change the float tensor
        parameters of the object (i.e. it preserves the state of the object).

        Parameters
        ----------
        method: Callable
            The method of this class to be tested

        Raises
        ------
        KeyError
            If the method does not preserve the float tensors of the object.

        """

        all_params0, names0 = _get_tensors(self)
        all_params0 = [p.clone() for p in all_params0]
        method(*args, **kwargs)
        all_params1, names1 = _get_tensors(self)

        # now assert if all_params0 == all_params1
        clsname = method.__self__.__class__.__name__
        methodname = method.__name__
        msg = "The method %s.%s does not preserve the object's float tensors: \n" % (
            clsname, methodname)
        if len(all_params0) != len(all_params1):
            msg += "The number of parameters changed:\n"
            msg += "* number of object's parameters before: %d\n" % len(
                all_params0)
            msg += "* number of object's parameters after : %d\n" % len(
                all_params1)
            raise KeyError(msg)

        for pname, p0, p1 in zip(names0, all_params0, all_params1):
            if p0.shape != p1.shape:
                msg += "The shape of %s changed\n" % pname
                msg += "* (before) %s.shape: %s\n" % (pname, p0.shape)
                msg += "* (after ) %s.shape: %s\n" % (pname, p1.shape)
                raise KeyError(msg)
            if not torch.allclose(p0, p1):
                msg += "The value of %s changed\n" % pname
                msg += "* (before) %s: %s\n" % (pname, p0)
                msg += "* (after ) %s: %s\n" % (pname, p1)
                raise KeyError(msg)

    def __assert_get_correct_params(self, method, *args, **kwargs):
        """This function perform checks if the getparams on the method returns
        the correct tensors

        Parameters
        ----------
        method: Callable
            The method of this class to be tested

        Raises
        ------
        KeyError
            If the method does not return the correct tensors.

        """

        methodname = method.__name__
        clsname = method.__self__.__class__.__name__

        # get all tensor parameters in the object
        all_params, all_names = _get_tensors(self)

        def _get_tensor_name(param):
            for i in range(len(all_params)):
                if id(all_params[i]) == id(param):
                    return all_names[i]
            return None

        # get the parameter tensors used in the operation and the tensors specified by the developer
        oper_names, oper_params = self.__list_operating_params(
            method, *args, **kwargs)
        user_names = self.getparamnames(method.__name__)
        user_params = [get_attr(self, name) for name in user_names]
        user_params_id = [id(p) for p in user_params]
        oper_params_id = [id(p) for p in oper_params]
        user_params_id_set = set(user_params_id)
        oper_params_id_set = set(oper_params_id)

        # check if the userparams contains non-tensor
        for i in range(len(user_params)):
            param = user_params[i]
            if (not isinstance(param, torch.Tensor)) or \
               (isinstance(param, torch.Tensor) and param.dtype not in torch_float_type):
                msg = "Parameter %s is a non-floating point tensor" % user_names[
                    i]
                raise KeyError(msg)

        # check if there are missing parameters (present in operating params, but not in the user params)
        missing_names = []
        for i in range(len(oper_names)):
            if oper_params_id[i] not in user_params_id_set:
                # if oper_names[i] not in user_names:
                missing_names.append(oper_names[i])
        # if there are missing parameters, give a warning (because the program
        # can still run correctly, e.g. missing parameters are parameters that
        # are never set to require grad)
        if len(missing_names) > 0:
            msg = "getparams for %s.%s does not include: %s" % (
                clsname, methodname, ", ".join(missing_names))
            warnings.warn(msg, stacklevel=3)

        # check if there are excessive parameters (present in the user params, but not in the operating params)
        excess_names = []
        for i in range(len(user_names)):
            if user_params_id[i] not in oper_params_id_set:
                # if user_names[i] not in oper_names:
                excess_names.append(user_names[i])
        # if there are excess parameters, give warnings
        if len(excess_names) > 0:
            msg = "getparams for %s.%s has excess parameters: %s" % \
                (clsname, methodname, ", ".join(excess_names))
            warnings.warn(msg, stacklevel=3)

    def __list_operating_params(self, method, *args, **kwargs):
        """Sequence the tensors used in executing the method by calling the method
        and see which parameters are connected in the backward graph"""

        # get all the tensors recursively
        all_tensors, all_names = _get_tensors(self)

        # copy the tensors and require them to be differentiable
        copy_tensors0 = [
            tensor.clone().detach().requires_grad_() for tensor in all_tensors
        ]
        copy_tensors = copy.copy(copy_tensors0)
        _set_tensors(self, copy_tensors)

        # run the method and see which one has the gradients
        output = method(*args, **kwargs)
        if not isinstance(output, torch.Tensor):
            raise RuntimeError(
                "The method to be asserted must have a tensor output")
        output = output.sum()
        grad_tensors = torch.autograd.grad(output,
                                           copy_tensors0,
                                           retain_graph=True,
                                           allow_unused=True)

        # return the original tensor
        all_tensors_copy = copy.copy(all_tensors)
        _set_tensors(self, all_tensors_copy)

        names = []
        params = []
        for i, grad in enumerate(grad_tensors):
            if grad is None:
                continue
            names.append(all_names[i])
            params.append(all_tensors[i])

        return names, params


# traversing functions
def _traverse_obj(obj: Any,
                  prefix: str,
                  action: Callable,
                  crit: Callable,
                  max_depth: int = 20,
                  exception_ids=None):
    """
    Traverse an object to get/set variables that are accessible through the object.
    The object can be a torch.nn.Module, a class instance, or an iterable object.
    The action is performed on the object that satisfies the criteria.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils.editable_module import _traverse_obj, torch_float_type
    >>> class A:
    ...     def __init__(self):
    ...         self.a = 2
    ...         self.b = torch.tensor(3.0)
    ...         self.c = torch.tensor(4.0)
    ...         self.d = torch.tensor(5.0)
    ...
    >>> a = A()
    >>> def action(elmt, name, objdict, key):
    ...     print(name, elmt)
    ...
    >>> def crit(elmt):
    ...     return isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type
    ...
    >>> _traverse_obj(a, "", action, crit)
    b tensor(3.)
    c tensor(4.)
    d tensor(5.)

    Parameters
    ----------
    obj: Any
        The object user wants to traverse down
    prefix: str
        Prefix of the name of the collected tensors.
    action: Callable
        The action to be performed on the object.
    crit: Callable
        The criteria to be met to perform the action.
    max_depth: int (default=20)
        Maximum recursive depth to avoid infinitely running program.
        If the maximum depth is reached, then raise a RecursionError.
    exception_ids: Set[int] (default=None)
        Set of ids of objects that are already traversed to avoid infinite loop.

    Raises
    ------
    RecursionError
        If the maximum depth is reached.
    RuntimeError
        If the object is not iterable or keyable.

    """
    if exception_ids is None:
        # None is set as default arg to avoid expanding list for multiple
        # invokes of _get_tensors without exception_ids argument
        exception_ids = set()

    if isinstance(obj, torch.nn.Module):
        generators = [obj._parameters.items(), obj._modules.items()]
        name_format = "{prefix}{key}"
        objdicts = [obj._parameters, obj._modules]
    elif hasattr(obj, "__dict__"):
        generators = [obj.__dict__.items()]
        name_format = "{prefix}{key}"
        objdicts = [obj.__dict__]
    elif hasattr(obj, "__iter__"):
        if isinstance(obj, dict):
            generators = [obj.items()]
        else:
            generators = [enumerate(obj)]  # type: ignore[list-item]
        name_format = "{prefix}[{key}]"
        objdicts = [obj]
    else:
        raise RuntimeError("The object must be iterable or keyable")

    for generator, objdict in zip(generators, objdicts):
        for key, elmt in generator:
            name = name_format.format(prefix=prefix, key=key)
            if crit(elmt):
                action(elmt, name, objdict, key)
                continue

            hasdict = hasattr(elmt, "__dict__")
            hasiter = hasattr(elmt, "__iter__")
            if hasdict or hasiter:
                # add exception to avoid infinite loop if there is a mutual dependant on objects
                if id(elmt) in exception_ids:
                    continue
                else:
                    exception_ids.add(id(elmt))

                if max_depth > 0:
                    _traverse_obj(elmt,
                                  action=action,
                                  crit=crit,
                                  prefix=name + "." if hasdict else name,
                                  max_depth=max_depth - 1,
                                  exception_ids=exception_ids)
                else:
                    raise RecursionError("Maximum number of recursion reached")


def _get_tensors(obj: Any, prefix="", max_depth=20):
    """
    Collect all tensors in an object recursively and return the tensors as well
    as their "names" (names meaning the address, e.g. "self.a[0].elmt").

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils.editable_module import _get_tensors
    >>> class A:
    ...     def __init__(self):
    ...         self.a = 2
    ...         self.b = torch.tensor(3.0)
    ...         self.c = torch.tensor(4.0)
    ...         self.d = torch.tensor(5.0)
    ...
    >>> a = A()
    >>> _get_tensors(a)
    ([tensor(3.), tensor(4.), tensor(5.)], ['b', 'c', 'd'])

    Parameters
    ----------
    obj: Any
        The object user wants to traverse down
    prefix: str (default="")
        Prefix of the name of the collected tensors.
    max_depth: int (default=20)
        Maximum recursive depth to avoid infinitely running program.
        If the maximum depth is reached, then raise a RecursionError.

    Returns
    -------
    res: list[torch.Tensor]
        Sequence of tensors collected recursively in the object.
    name: list[str]
        Sequence of names of the collected tensors.

    """

    # get the tensors recursively towards torch.nn.Module
    res = []
    names = []

    def action(elmt, name, objdict, key):
        res.append(elmt)
        names.append(name)

    # traverse down the object to collect the tensors
    def crit(elmt):
        return isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type

    _traverse_obj(obj,
                  action=action,
                  crit=crit,
                  prefix=prefix,
                  max_depth=max_depth)
    return res, names


def _set_tensors(obj: Any, all_params: List[torch.Tensor], max_depth: int = 20):
    """
    Set the tensors in an object to new tensor object listed in `all_params`.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils.editable_module import _set_tensors
    >>> class A:
    ...     def __init__(self):
    ...         self.a = 2
    ...         self.b = torch.tensor(3.0)
    ...         self.c = torch.tensor(4.0)
    ...         self.d = torch.tensor(5.0)
    ...
    >>> a = A()
    >>> _set_tensors(a, [torch.tensor(6.0), torch.tensor(7.0), torch.tensor(8.0)])
    >>> a.b
    tensor(6.)
    >>> a.c
    tensor(7.)

    Parameters
    ----------
    obj: an instance
        The object user wants to traverse down
    all_params: List[torch.Tensor]
        Sequence of tensors to be put in the object.
    max_depth: int (default=20)
        Maximum recursive depth to avoid infinitely running program.
        If the maximum depth is reached, then raise a RecursionError.

    """

    def action(elmt, name, objdict, key):
        objdict[key] = all_params.pop(0)

    # traverse down the object to collect the tensors
    def crit(elmt):
        return isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type

    _traverse_obj(obj, action=action, crit=crit, prefix="", max_depth=max_depth)



"""
Density Functional Theory.
"""
# flake8: noqa
import logging

logger_ = logging.getLogger(__name__)

try:
    from deepchem.models.dft.hamilton.orbparams import BaseOrbParams
    from deepchem.models.dft.hamilton.orbparams import QROrbParams
    from deepchem.models.dft.hamilton.orbparams import MatExpOrbParams
except ModuleNotFoundError as e:
    logger_.warning(
        f'Skipped loading some Pytorch utilities, missing a dependency. {e}')



"""
qm8 dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

GDB8_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb8.tar.gz"
QM8_CSV_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm8.csv"
QM8_TASKS = [
    "E1-CC2", "E2-CC2", "f1-CC2", "f2-CC2", "E1-PBE0", "E2-PBE0", "f1-PBE0",
    "f2-PBE0", "E1-PBE0", "E2-PBE0", "f1-PBE0", "f2-PBE0", "E1-CAM", "E2-CAM",
    "f1-CAM", "f2-CAM"
]


class _QM8Loader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "qm8.sdf")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=GDB8_URL,
                                             dest_dir=self.data_dir)
            dc.utils.data_utils.untargz_file(
                os.path.join(self.data_dir, "gdb8.tar.gz"), self.data_dir)
        loader = dc.data.SDFLoader(tasks=self.tasks,
                                   featurizer=self.featurizer,
                                   sanitize=True)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_qm8(
    featurizer: Union[dc.feat.Featurizer, str] = dc.feat.CoulombMatrix(26),
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load QM8 dataset

    QM8 is the dataset used in a study on modeling quantum
    mechanical calculations of electronic spectra and excited
    state energy of small molecules. Multiple methods, including
    time-dependent density functional theories (TDDFT) and
    second-order approximate coupled-cluster (CC2), are applied to
    a collection of molecules that include up to eight heavy atoms
    (also a subset of the GDB-17 database). In our collection,
    there are four excited state properties calculated by four
    different methods on 22 thousand samples:

    S0 -> S1 transition energy E1 and the corresponding oscillator strength f1

    S0 -> S2 transition energy E2 and the corresponding oscillator strength f2

    E1, E2, f1, f2 are in atomic units. f1, f2 are in length representation

    Random splitting is recommended for this dataset.

    The source data contain:

    - qm8.sdf: molecular structures
    - qm8.sdf.csv: tables for molecular properties

    - Column 1: Molecule ID (gdb9 index) mapping to the .sdf file
    - Columns 2-5: RI-CC2/def2TZVP
    - Columns 6-9: LR-TDPBE0/def2SVP
    - Columns 10-13: LR-TDPBE0/def2TZVP
    - Columns 14-17: LR-TDCAM-B3LYP/def2TZVP

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Note
    ----
    DeepChem 2.4.0 has turned on sanitization for this dataset by
    default.  For the QM8 dataset, this means that calling this
    function will return 21747 compounds instead of 21786 in the source
    dataset file.  This appears to be due to valence specification
    mismatches in the dataset that weren't caught in earlier more lax
    versions of RDKit.  Note that this may subtly affect benchmarking
    results on this dataset.

    References
    ----------
    .. [1] Blum, Lorenz C., and Jean-Louis Reymond. "970 million druglike
        small molecules for virtual screening in the chemical universe database
        GDB-13." Journal of the American Chemical Society 131.25 (2009):
        8732-8733.
    .. [2] Ramakrishnan, Raghunathan, et al. "Electronic spectra from TDDFT
        and machine learning in chemical space." The Journal of chemical physics
        143.8 (2015): 084111.
    """
    loader = _QM8Loader(featurizer, splitter, transformers, QM8_TASKS, data_dir,
                        save_dir, **kwargs)
    return loader.load_dataset('qm8', reload)



"""
Geometric utility functions for 3D geometry.
"""
import numpy as np
from scipy.spatial.distance import cdist
from copy import deepcopy


def unit_vector(vector: np.ndarray) -> np.ndarray:
    """ Returns the unit vector of the vector.

    Parameters
    ----------
    vector: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).

    Returns
    ----------
    np.ndarray
        A numpy array of shape `(3,)`. The unit vector of the input vector.
    """
    return vector / np.linalg.norm(vector)


def angle_between(vector_i: np.ndarray, vector_j: np.ndarray) -> float:
    """Returns the angle in radians between vectors "vector_i" and "vector_j"

    Note that this function always returns the smaller of the two angles between
    the vectors (value between 0 and pi).

    Parameters
    ----------
    vector_i: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).
    vector_j: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).

    Returns
    ----------
    np.ndarray
        The angle in radians between the two vectors.

    Examples
    --------
    >>> print("%0.06f" % angle_between((1, 0, 0), (0, 1, 0)))
    1.570796
    >>> print("%0.06f" % angle_between((1, 0, 0), (1, 0, 0)))
    0.000000
    >>> print("%0.06f" % angle_between((1, 0, 0), (-1, 0, 0)))
    3.141593
    """
    vector_i_u = unit_vector(vector_i)
    vector_j_u = unit_vector(vector_j)
    angle = np.arccos(np.dot(vector_i_u, vector_j_u))
    if np.isnan(angle):
        if np.allclose(vector_i_u, vector_j_u):
            return 0.0
        else:
            return np.pi
    return angle


def generate_random_unit_vector() -> np.ndarray:
    r"""Generate a random unit vector on the sphere S^2.

    Citation: http://mathworld.wolfram.com/SpherePointPicking.html

    Pseudocode:
      a. Choose random theta \element [0, 2*pi]
      b. Choose random z \element [-1, 1]
      c. Compute output vector u: (x,y,z) = (sqrt(1-z^2)*cos(theta), sqrt(1-z^2)*sin(theta),z)

    Returns
    -------
    u: np.ndarray
        A numpy array of shape `(3,)`. u is an unit vector
    """
    theta = np.random.uniform(low=0.0, high=2 * np.pi)
    z = np.random.uniform(low=-1.0, high=1.0)
    u = np.array([
        np.sqrt(1 - z**2) * np.cos(theta),
        np.sqrt(1 - z**2) * np.sin(theta), z
    ])
    return u


def generate_random_rotation_matrix() -> np.ndarray:
    r"""Generates a random rotation matrix.

    1. Generate a random unit vector u, randomly sampled from the
      unit sphere (see function generate_random_unit_vector()
      for details)
    2. Generate a second random unit vector v
      a. If absolute value of u \dot v > 0.99, repeat.
        (This is important for numerical stability. Intuition: we
        want them to be as linearly independent as possible or
        else the orthogonalized version of v will be much shorter
        in magnitude compared to u. I assume in Stack they took
        this from Gram-Schmidt orthogonalization?)
      b. v" = v - (u \dot v)*u, i.e. subtract out the component of
        v that's in u's direction
      c. normalize v" (this isn"t in Stack but I assume it must be
        done)

    3. find w = u \cross v"
    4. u, v", and w will form the columns of a rotation matrix, R.
      The intuition is that u, v" and w are, respectively, what
      the standard basis vectors e1, e2, and e3 will be mapped
      to under the transformation.

    Returns
    -------
    R: np.ndarray
        A numpy array of shape `(3, 3)`. R is a rotation matrix.
    """
    u = generate_random_unit_vector()
    v = generate_random_unit_vector()
    while np.abs(np.dot(u, v)) >= 0.99:
        v = generate_random_unit_vector()

    vp = v - (np.dot(u, v) * u)
    vp /= np.linalg.norm(vp)
    w = np.cross(u, vp)
    R = np.column_stack((u, vp, w))
    return R


def rotate_molecules(mol_coordinates_list):
    """Rotates provided molecular coordinates.

    Pseudocode:
    1. Generate random rotation matrix. This matrix applies a random
      transformation to any 3-vector such that, were the random transformation
      repeatedly applied, it would randomly sample along the surface of a sphere
      with radius equal to the norm of the given 3-vector cf.
      _generate_random_rotation_matrix() for details
    2. Apply R to all atomic coordinatse.
    3. Return rotated molecule
    """
    R = generate_random_rotation_matrix()
    rotated_coordinates_list = []

    for mol_coordinates in mol_coordinates_list:
        coordinates = deepcopy(mol_coordinates)
        rotated_coordinates = np.transpose(np.dot(R, np.transpose(coordinates)))
        rotated_coordinates_list.append(rotated_coordinates)

    return (rotated_coordinates_list)


def is_angle_within_cutoff(vector_i: np.ndarray, vector_j: np.ndarray,
                           angle_cutoff: float) -> bool:
    """A utility function to compute whether two vectors are within a cutoff from 180 degrees apart.

    Parameters
    ----------
    vector_i: np.ndarray
        A numpy array of shape (3,)`, where `3` is (x,y,z).
    vector_j: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).
    cutoff: float
        The deviation from 180 (in degrees)

    Returns
    -------
    bool
        Whether two vectors are within a cutoff from 180 degrees apart
    """
    angle = angle_between(vector_i, vector_j) * 180. / np.pi
    return (angle > (180 - angle_cutoff) and angle < (180. + angle_cutoff))


def compute_centroid(coordinates: np.ndarray) -> np.ndarray:
    """Compute the (x,y,z) centroid of provided coordinates

    Parameters
    ----------
    coordinates: np.ndarray
        A numpy array of shape `(N, 3)`, where `N` is the number of atoms.

    Returns
    -------
    centroid: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).
    """
    centroid = np.mean(coordinates, axis=0)
    return centroid


def compute_protein_range(coordinates: np.ndarray) -> np.ndarray:
    """Compute the protein range of provided coordinates

    Parameters
    ----------
    coordinates: np.ndarray
        A numpy array of shape `(N, 3)`, where `N` is the number of atoms.

    Returns
    -------
    protein_range: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).
    """
    protein_max = np.max(coordinates, axis=0)
    protein_min = np.min(coordinates, axis=0)
    protein_range = protein_max - protein_min
    return protein_range


def subtract_centroid(coordinates: np.ndarray,
                      centroid: np.ndarray) -> np.ndarray:
    """Subtracts centroid from each coordinate.

    Subtracts the centroid, a numpy array of dim 3, from all coordinates
    of all atoms in the molecule

    Note that this update is made in place to the array it's applied to.

    Parameters
    ----------
    coordinates: np.ndarray
        A numpy array of shape `(N, 3)`, where `N` is the number of atoms.
    centroid: np.ndarray
        A numpy array of shape `(3,)`

    Returns
    -------
    coordinates: np.ndarray
        A numpy array of shape `(3,)`, where `3` is (x,y,z).
    """
    coordinates -= np.transpose(centroid)
    return coordinates


def compute_pairwise_distances(first_coordinate: np.ndarray,
                               second_coordinate: np.ndarray) -> np.ndarray:
    """Computes pairwise distances between two molecules.

    Takes an input (m, 3) and (n, 3) numpy arrays of 3D coords of
    two molecules respectively, and outputs an m x n numpy
    array of pairwise distances in Angstroms between the first and
    second molecule. entry (i,j) is dist between the i"th
    atom of first molecule and the j"th atom of second molecule.

    Parameters
    ----------
    first_coordinate: np.ndarray
        A numpy array of shape `(m, 3)`, where `m` is the number of atoms.
    second_coordinate: np.ndarray
        A numpy array of shape `(n, 3)`, where `n` is the number of atoms.

    Returns
    -------
    pairwise_distances: np.ndarray
        A numpy array of shape `(m, n)`
    """

    pairwise_distances = cdist(first_coordinate,
                               second_coordinate,
                               metric='euclidean')
    return pairwise_distances
